/*  Demonstration for the 3x+1 problem, also called the "Syracuse problem"  */
/*  File:  3x+1.cs							    */

#include <stdio.h>
#include <stdlib.h>

#define NC 16	/* Number of classes explored in parallel */
			/* 16 classes is suitable for simulation  */

shape [NC]classes;	/* Classes of numbers == 3 (mod 4)	*/

unsigned long:classes	X0,	/* Starting value for current sequence	*/
			X;	/* Current sequence value		*/

unsigned int:classes	C,	/* Number of steps in current sequence	*/
			Cmax;	/* Maximun number of steps to stopping	*/

unsigned int Ctot;		/* Total steps since starting		*/

unsigned long step;		/* Amount to advance X0 for next seq.	*/


/*  Initialize search	*/

void start()
 { with (classes)
    { X0 = 4*pcoord(0) + 3;	/* Smallest number in class */
      X = X0;  C = Cmax = 0;
      step = 4*NC;  Ctot = 0;
      printf("Initialized with %d classes, step = %d\n", NC, step);
    }
 }


/*  Report all positions at which X == X0.  It is extremely unlikely	*/
/*  that this will ever be used, except for testing.			*/
/*  Call with classes selected, vp's with cycles active (or more).	*/

void report_cycles()
 { unsigned int:classes i;	/* Processor coordinates */
   bool:classes cy;   unsigned int p;
   printf("CYCLE(S) DETECTED AT Processor      X0          C\b\n");
   i = pcoord(0);
   cy = (X == X0);		/* True where sequence cycle detected */
   while ( p = (<?= (cy ? i : NC)),	/* Lowest cycle position */
			    p < NC )	/* if any		 */
    { printf("                      %7d   %10d  %10d\n",
				     p,  [p]X0, [p]C    );
      [p]cy = 0;	/* This item reported */
    }
   printf("\b\n");
  }


/*  Artificial test of report_cycles.  The program is careful to avoid	*/
/*  any net change of the computation state.				*/

void testrc()
 { unsigned long:classes Xsv;	/* To save X */
   printf("Test of cycle report -- odd vp's should report cycles.\n");
   with (classes) everywhere
    { Xsv = X;  X = X0 + 1;
      where ( pcoord(0)&1 ) { X = X0;  report_cycles(); }
      X = Xsv;			/* So X isn't ultimately changed */
    }
 }


/*  Perform n steps of calculation.				*/
/*  Should be called with classes selected, all vp's active.	*/

void stepper(unsigned int n)
 { unsigned int k;		/* Loop counter			*/
   unsigned long:classes Y;	/* Working parallel variable	*/
   everywhere			/* May help compiler	*/
   for ( k = 0;  k < n;  k++ )
    { Y = X>>1;			/* X div 2 (integer)	*/
      where ( X&1 )		/* where X is odd	*/
       X = 1 + X + Y;		/* (3*x + 1)/2		*/
      else
       X = Y;			/* X/2			*/
      C++;			/* Count step		*/
      where ( X == X0 )		/* Cycle in sequence	*/
       if ( |= (int:classes) 1 ) report_cycles();
      where ( X < X0 )		/* Seqence stops - new sequence	*/
       { X0 = X0 + step;	/* New start	*/
	 X = X0;
	 Cmax >?= C;		/* Retain longest sequence	*/
	 C = 0;			/* Reset to count next sequence	*/
       }
    }	/* end for	*/
 }	/* end stepper(n)	*/


/*  Run and time n calculation cycles, report results.	*/

void run(int n)
 { with (classes) everywhere
    { CM_timer_clear(1);  CM_timer_start(1);
      stepper(n);
      CM_timer_stop(1);
      Ctot += n;
      printf("Confirmed to %d,  longest sequence %d,  %d total steps.",
	             (<?= X0) - 1,         (>?= Cmax), Ctot              );
      CM_timer_print(1);
    }
 }


/*  Report complete state of computation.  Works in any situation, but	*/
/*  intended for use with small-scale test calculations.		*/

void report_state()
 { int p;
   printf("Processor      X0           X         Cmax          C\n");
   for ( p = 0;  p < NC;  p++ )
    printf("%8d   %10d  %10d  %10d  %10d\n",
	      p, [p]X0, [p]X, [p]Cmax, [p]C);
 }


/*  Main program which manages interactive experiments	*/
/*  Commands are:					*/
/*		  r <n>	- run n steps			*/
/*		  i	- (re)initialize calculation	*/
/*		  s	- report complete state		*/
/*		  t	- test cycle report		*/
/*		  q	- quit				*/

void main()
 { int ns;  char buff[80];
   printf("3x + 1 calculation using %d virtual processors.\n", NC);
   while ( printf("3x+1> "),  scanf("%s", buff),  buff[0] != 'q' )
    switch ( buff[0] )
     { case 'r':  scanf("%d", &ns);  
		  if ( [0]X0&1 ) run(ns);
		    else
		     printf("Calculation isn't initialized properly.\n");
		  break;
       case 'i':  start(); break;
       case 's':  report_state(); break;
       case 't':  testrc(); break;
       default:   printf("Bad input -- ignored.\n"); break;
     }
 }
